from typing import Optional

from pydantic import BaseModel

from parrot1.commons.models.language import Language


class ASRConfigs(BaseModel):
    model_type_or_size: str
    temperature: Optional[float] = 0.1
    prompt: Optional[str] = ""
    beam_size: Optional[int] = 5

    class Config:
        frozen: True


class LLMConfigs(BaseModel):
    model_type_or_size: str
    repo_id: Optional[str] = None

    class Config:
        frozen: True


class LLMParamsConfigs(BaseModel):
    temperature: Optional[float] = 0.2
    top_k: Optional[int] = 40
    top_p: Optional[float] = 0.9
    max_tokens: Optional[int] = 4096

    class Config:
        frozen: True


class ASRModelsConfigs(BaseModel):
    whisper: ASRConfigs
    faster_whisper: ASRConfigs

    class Config:
        frozen: True


class GenerativeModelsConfigs(BaseModel):
    chunking: Optional[LLMParamsConfigs]
    text_generation: Optional[LLMParamsConfigs]
    openai: LLMConfigs
    llama_cpp: LLMConfigs

    class Config:
        frozen: True


class ParrotConfigs(BaseModel):
    language: Optional[Language] = Language.IT.value
    asr_models: ASRModelsConfigs
    generative_models: GenerativeModelsConfigs

    class Config:
        frozen: True



import logging
import os
from typing import Optional, Any, List

from pydub import AudioSegment
from tqdm import tqdm
from parrot1 import PARROT_CACHED_MODELS
from parrot1.audio.transcription.model import TimedTranscription, TimedPiece
from parrot1.commons.asr.base import BaseASRModel
from parrot1.commons.asr.utils import format_audio_onto_ndarray
from parrot1.commons.models.language import Language

try:
    from faster_whisper import WhisperModel

    has_faster_whisper = True
except ImportError:
    has_faster_whisper = False


class FasterWhisper(BaseASRModel):
    def __init__(
        self,
        model_size_or_type: str,
        language: Optional[Language] = Language.IT,
        prompt: Optional[str] = "",
        temperature: Optional[float] = 0.2,
        beam_size: Optional[int] = 5,
        client_or_model: Any = None,
        download_root: os.PathLike = None,
    ):
        super().__init__(
            model_size_or_type,
            language,
            prompt,
            temperature,
            beam_size,
            client_or_model,
        )
        self.__logger = logging.getLogger("FasterWhisper")
        if self.client_or_model is None:
            if has_faster_whisper:
                self.client_or_model = WhisperModel(
                    model_size_or_path=self.model_size_or_type,
                    download_root=download_root or PARROT_CACHED_MODELS,
                )
            else:
                self.__logger.error(
                    "The faster-whisper package was not installed. Try fixing it by doing pip install parrot1[faster-whisper]."
                    "If not present, you cannot use a free ASR model such Whisper. "
                    "Alternatively, set the OPENAI_API_KEY environment variable."
                )

    async def atranscribe(
        self, audio: AudioSegment, vtt: bool = False
    ) -> TimedTranscription:
        pass

    def transcribe(self, audio: AudioSegment, vtt: bool = False) -> TimedTranscription:
        audio_array = format_audio_onto_ndarray(audio)

        segments, _ = self.client_or_model.transcribe(
            audio_array,
            language=self.language,
            beam_size=self.beam_size,
            temperature=self.temperature,
            initial_prompt=self.prompt,
        )

        segments = list(segments)

        text = " ".join(s.text for s in segments)
        pieces = []
        if vtt:
            pieces = [
                TimedPiece(transcript=s.text, start=s.start, end=s.end)
                for s in segments
            ]
        return TimedTranscription(text=text, pieces=pieces)

    def transcribe_chunks(
        self, audio_chunks: List[AudioSegment], vtt: bool = False
    ) -> List[TimedTranscription]:
        transcription_chunks = [
            self.transcribe(ac, vtt=vtt) for ac in tqdm(audio_chunks)
        ]

        return transcription_chunks



import importlib.util
import logging
import os.path
from typing import Union, List

from parrot1.audio.transcription.model import TimedTranscription

from parrot1 import PARROT_CACHED_MODELS
from parrot1.commons.asr.base import BaseASRModel
from parrot1.commons.asr.faster_whisper import FasterWhisper
from parrot1.commons.asr.openai_whisper import OpenaiWhisper
from parrot1.config.config import PARROT_CONFIGS

from parrot1.audio.extraction.audio_extraction import (
    get_audio_from_video,
    split_audio_for_size,
)

__logger = logging.getLogger(__file__)

s = importlib.util.find_spec(name="faster_whisper")

has_faster_whisper = s is not None


def get_client(use_faster_whisper: bool = False) -> Union[BaseASRModel, None]:
    if os.getenv("OPENAI_API_KEY") is None and not use_faster_whisper:
        __logger.warning(
            "OPENAI_API_KEY is not set but you're trying to use the OpenAI Apis."
            "The model choice will fallback to faster-whisper automatically."
        )
    if os.getenv("OPENAI_API_KEY") is not None and not use_faster_whisper:
        return OpenaiWhisper(
            model_size_or_type=PARROT_CONFIGS.asr_models.whisper.model_type_or_size,
            language=PARROT_CONFIGS.language.value,
            temperature=PARROT_CONFIGS.asr_models.whisper.temperature,
            prompt=PARROT_CONFIGS.asr_models.whisper.prompt,
        )
    elif has_faster_whisper:
        cache_root = PARROT_CACHED_MODELS
        __logger.info(
            "Using Faster Whisper model."
            f"Using model cached at folder at {cache_root.as_posix()}"
        )
        os.makedirs(cache_root, exist_ok=True)
        return FasterWhisper(
            model_size_or_type=PARROT_CONFIGS.asr_models.faster_whisper.model_type_or_size,
            language=PARROT_CONFIGS.language.value,
            prompt=PARROT_CONFIGS.asr_models.faster_whisper.prompt,
            temperature=PARROT_CONFIGS.asr_models.faster_whisper.temperature,
            beam_size=PARROT_CONFIGS.asr_models.faster_whisper.beam_size,
            download_root=cache_root,
        )
    else:
        __logger.error(
            "The faster-whisper package was not installed. Try fixing it by doing pip install parrot1[faster-whisper]."
            "If not present, you cannot use a free ASR model such Whisper. "
            "Please set the OPENAI_API_KEY environment variable."
        )
        return None


def transcribe_video_source(
    filepath: Union[str, os.PathLike],
    max_time: int = 3 * 60,
    transcript: str = None,
    vtt: bool = False,
    use_faster_whisper: bool = False,
) -> List[TimedTranscription]:
    __logger.info(f"Transcribing video source at {filepath}")
    audio = get_audio_from_video(video_filename=filepath)
    audio_chunks = split_audio_for_size(audio, max_time=max_time)

    # Get the right model
    client = get_client(use_faster_whisper)

    if client is not None:
        transcription_chunks = client.transcribe_chunks(audio_chunks, vtt)
    else:
        raise RuntimeError("No available method for ASR is possible.")

    return transcription_chunks



## PROMPT Storage for different languages (actually english and italian)

# ITALIAN

CHUNK_PROMPT_IT = """
Genera un riassunto di massimo 100 parole data la trascrizione di una chiamata su Microsoft Teams che ti verrà fornita di seguito.
Il riassunto deve essere una descrizione dei punti più importanti menzionati nella trascrizione.
Devi identificare i punti più importanti e riportarli con delle frasi autoconclusive e concise.
Rifletti sul tuo task punto per punto e genera il riassunto.

Testo:
{text}

Riassunto:
"""


EMAIL_PROMPT_IT = """
Sei un PM incaricato di generare delle email di recap che siano consistenti e ben fatte.
Genera una email di riassunto partendo dalla trascrizione testuale di una chiamata su Microsoft Teams. 
La mail deve essere di massimo 300 parole e devi selezionare i punti più importanti dati i seguenti testi.
Nella mail scrivi inizialmente un piccolo riassunto del tema principale, inserendo successivamente un elenco puntato con al massimo 10 punti salienti per la trascrizione.
Puoi anche usare meno di 10 punti per l'elenco puntato, ma non utilizzarne più di 10.
Concludi la mail con una frase di congedo. La mail deve necessariamente concludersi senza interruzioni. 
Deve contenere massimo 300 parole.

Trascrizione:
{texts}

E-mail:
"""


REPORT_PROMPT_IT = """
Sulla base delle informazioni contenute nella trascrizione seguente, genera un report.
Il report dovrebbe fornire un'analisi dettagliata di ciascun argomento discusso, spiegando come ciascun argomento ha contribuito alla riunione.
Concentrati sulla pertinenza, l’affidabilità e il significato di ciascun argomento.
Assicurati che il report sia ben strutturato, informativo, approfondito e segua una sintassi Markdown.
Includi fatti, domande e numeri rilevanti quando disponibili.
Il report dovrà avere una lunghezza minima di 1.200 parole e di massimo 2,500 parole.

Trascrizione:
{texts}

Report:
"""

# ENGLISH

CHUNK_PROMPT_EN = """
Generate a summary of up to 100 words given the Microsoft Teams call transcript provided below.
The summary should be a description of the most important points mentioned in the transcript.
You must identify the most important points and report them in self-contained and concise sentences.
Reflect on your task point by point and generate the summary.

Text:
{text}

Summary:
"""


EMAIL_PROMPT_EN = """
You are a PM tasked with generating delivery emails that are consistent and well-crafted.
Generate a summary email starting from the text transcription of a call on Microsoft Teams.
The email must be a maximum of 300 words and you must select the most important points given the following texts.
In the email, initially write a small summary of the main theme, subsequently inserting a bulleted list with a maximum of 10 salient points for transcription.
You can also use fewer than 10 bullet points, but don't use more than 10.
Conclude the email with a parting sentence. The email must necessarily end without interruption.
Must contain a maximum of 300 words.

Transcription:
{texts}

E-mail:
"""


REPORT_PROMPT_EN = """
Based on the information in the following transcription, generate a report. 
The report should provide a detailed analysis of each discussed topic, explaining how each topic contributed to the meeting.
Focus on the relevance, reliability, and significance of each topic.
Ensure that the report is well-structured, informative, in-depth, and follows Markdown syntax.
Include relevant facts, questions, and numbers whenever available.
The report should have a minimum length of 1,200 words and maximum of 2,500.

Transcription:
{texts}

Report:
"""



import logging
import os
from typing import List, Union

from pydub import AudioSegment
from tqdm import tqdm

from parrot1.utils.file_utils import get_extension
from parrot1.audio.utils.silence import split_on_silence


__logger = logging.getLogger(__name__)


def get_audio_from_video(video_filename: Union[str, os.PathLike]) -> AudioSegment:
    """
    Takes the audio from the video file
    :param video_filename: (Union[str, os.PathLike]) path to the video
    :return: (io.BytesIO) Audio bytes
    """
    if not os.path.exists(video_filename):
        raise FileNotFoundError(f"File at {video_filename} does not exists.")
    audio = AudioSegment.from_file(video_filename, format=get_extension(video_filename))
    return audio


def split_audio_for_size(audio: AudioSegment, max_time: int = 60) -> List[AudioSegment]:
    """
    Splits the video in chunks, in order to retrieve the audio and feed it to OpenAI Whisper.
    :param max_time: The maximum time of a chunk. Used as an upper bound for reassembling smaller chunks.
    :param audio: The AudioSegment object to be splitted.
    :return: a List of AudioSegment representing the chunks of the recording.
    """
    max_millis = max_time * 1000
    __logger.debug(f"Using max chunk time {max_millis}")
    # Splits on silence
    audio_chunks, _ = split_on_silence(
        audio, min_silence_len=1000, silence_thresh=-40, padding=100
    )

    # Reassemble the split chunks on the max length defined
    __logger.info("Generating Audio chunks...")
    reassembled_chunks = [audio_chunks[0]]
    for chunk in tqdm(audio_chunks[1:]):
        if len(reassembled_chunks[-1]) < max_millis:
            reassembled_chunks[-1] += chunk
        else:
            reassembled_chunks.append(chunk)
    __logger.info(
        f"Finished chunking audio. Obtained {len(reassembled_chunks)} from a recording "
        f"of {len(audio)//1000} seconds"
    )

    return reassembled_chunks



from abc import ABC, abstractmethod
from typing import Any, List, Optional


class BaseLLMModel(ABC):
    def __init__(
        self,
        model_size_or_type: str,
        client_or_model: Optional[Any] = None,
    ):
        super().__init__()
        self.model_size_or_type = model_size_or_type
        self.client_or_model = client_or_model

    @abstractmethod
    async def agenerate(self, prompt: str, **kwargs) -> str:
        raise NotImplementedError("Abstract class. You should reimplement the method.")

    @abstractmethod
    def generate(self, prompt: str, **kwargs) -> str:
        raise NotImplementedError("Abstract class. You should reimplement the method.")

    @abstractmethod
    async def generate_from_prompts(self, prompts: List[str], **kwargs) -> List[str]:
        raise NotImplementedError("Abstract class. You should reimplement the method.")

    @abstractmethod
    def __call__(self, prompt: str, **kwargs) -> str:
        raise NotImplementedError("Abstract class. You should reimplement the method.")



import logging
import os.path
from typing import Any, List, Optional

try:
    import huggingface_hub
    from llama_cpp import Llama

    has_llama_cpp_extra = True
except ImportError:
    has_llama_cpp_extra = False
from tqdm import tqdm

from parrot1 import PARROT_CACHED_MODELS
from parrot1.commons.generative.base import BaseLLMModel


class LlamaCppModel(BaseLLMModel):
    def __init__(
        self,
        repo_id: str,
        model_size_or_type: str,
        client_or_model: Optional[Any] = None,
    ):
        super().__init__(model_size_or_type, client_or_model)
        self.__logger = logging.getLogger("LlamaCpp")
        self.repo_id = repo_id
        if self.client_or_model is None:
            # Download from hugging face
            model_path = PARROT_CACHED_MODELS / self.model_size_or_type
            if not os.path.exists(model_path):
                huggingface_hub.hf_hub_download(
                    repo_id=self.repo_id,
                    filename=self.model_size_or_type,
                    local_dir=PARROT_CACHED_MODELS,
                    local_dir_use_symlinks=False,
                )
            self.__logger.info(f"Loading model from {model_path}")
            self.client_or_model = Llama(
                model_path=model_path.as_posix(), verbose=False, n_ctx=4096
            )

    async def agenerate(self, prompt: str, **kwargs) -> str:
        return self.client_or_model.create_completion(prompt=prompt, **kwargs)[
            "choices"
        ][0]["text"]

    def generate(self, prompt: str, **kwargs) -> str:
        return self.client_or_model.create_completion(prompt=prompt, **kwargs)[
            "choices"
        ][0]["text"]

    def __call__(self, prompt: str, **kwargs) -> str:
        return self.generate(prompt, **kwargs)

    async def generate_from_prompts(self, prompts: List[str], **kwargs) -> List[str]:
        return [self.generate(prompt, **kwargs) for prompt in tqdm(prompts)]



import json
import logging

from pydantic.error_wrappers import ValidationError

from parrot1 import PARROT_CONFIG_FILE
from parrot1.config.config_model import ParrotConfigs

__logger = logging.getLogger(__name__)


def load_configurations():
    with open(PARROT_CONFIG_FILE, "r") as config_file:
        configs = ParrotConfigs(**json.load(config_file))
    return configs


PARROT_CONFIGS = None
try:
    PARROT_CONFIGS = load_configurations()
except ValidationError as exc:
    __logger.error(
        "Something bad happened during the configurations loading."
        "Try running the 'parrot1 reload-configs' command in order to reset the configurations to a correct formatting."
    )
    __logger.error(exc)



from abc import ABC, abstractmethod
from typing import Optional, Any, List

from pydub import AudioSegment

from parrot1.audio.transcription.model import TimedTranscription
from parrot1.commons.models.language import Language


class BaseASRModel(ABC):
    def __init__(
        self,
        model_size_or_type: str,
        language: Optional[Language] = Language.IT,
        prompt: Optional[str] = "",
        temperature: Optional[float] = 0.2,
        beam_size: Optional[int] = 5,
        client_or_model: Any = None,
    ):
        super().__init__()
        self.model_size_or_type = model_size_or_type
        self.language = language
        self.prompt = prompt
        self.temperature = temperature
        self.beam_size = beam_size
        self.client_or_model = client_or_model

    @abstractmethod
    async def atranscribe(
        self, audio: AudioSegment, vtt: bool = False
    ) -> TimedTranscription:
        raise NotImplementedError("Abstract class. You should reimplement the method.")

    @abstractmethod
    def transcribe(self, audio: AudioSegment, vtt: bool = False) -> TimedTranscription:
        raise NotImplementedError("Abstract class. You should reimplement the method.")

    @abstractmethod
    def transcribe_chunks(
        self, audio_chunks: List[AudioSegment], vtt: bool = False
    ) -> List[TimedTranscription]:
        raise NotImplementedError("Abstract class. You should reimplement the method.")



import os
import pathlib as pl
from typing import Union


def change_extension(filename: Union[str, os.PathLike], new_extension: str) -> pl.Path:
    return pl.Path(f"{filename.rsplit('.', 1)[0]}.{new_extension}")


def get_extension(filename: Union[str, os.PathLike]) -> str:
    return os.path.basename(filename).rsplit(".", 1)[1]


def get_filename(filepath: Union[str, os.PathLike]) -> str:
    return os.path.basename(filepath).rsplit(".", 1)[0]



import logging
import importlib.util
import os
from typing import List, Union

from parrot1 import PARROT_CACHED_MODELS
from parrot1.audio.transcription.model import TimedTranscription

from parrot1.commons.generative.base import BaseLLMModel
from parrot1.commons.generative.llamacpp import LlamaCppModel
from parrot1.commons.generative.openai_gpt import OpenaiGPTModel
from parrot1.config.config import PARROT_CONFIGS

from parrot1.recap.tasks import ParrotTask, resolve_prompt_from_task

imp_llama_cpp = importlib.util.find_spec(name="llama_cpp")

has_llama_cpp = imp_llama_cpp is not None

__logger = logging.getLogger(__name__)


def get_client(use_llama_cpp: bool = False) -> Union[BaseLLMModel, None]:
    if not use_llama_cpp:
        if os.getenv("OPENAI_API_KEY") is not None:
            return OpenaiGPTModel(
                model_size_or_type=PARROT_CONFIGS.generative_models.openai.model_type_or_size
            )
        else:
            __logger.error(
                "OPENAI_API_KEY is not set but you're trying to use the OpenAI Apis."
            )
            return None
    else:
        if has_llama_cpp:
            cache_root = PARROT_CACHED_MODELS
            __logger.info("Using llama_cpp model")
            __logger.info(f"Using cache folder at {cache_root.as_posix()}")
            os.makedirs(cache_root, exist_ok=True)
            return LlamaCppModel(
                repo_id=PARROT_CONFIGS.generative_models.llama_cpp.repo_id,
                model_size_or_type=PARROT_CONFIGS.generative_models.llama_cpp.model_type_or_size,
            )
        else:
            __logger.error(
                "The llama-cpp-python package was not installed. Try fixing it by doing pip install parrot1[llama-cpp]."
            )
            return None


async def generate_chunks(client: BaseLLMModel, texts: List[str]) -> List[str]:
    prompt = resolve_prompt_from_task(
        ParrotTask.CHUNK, language=PARROT_CONFIGS.language
    )

    summaries = await client.generate_from_prompts(
        prompts=[prompt.format(text=text) for text in texts],
        max_tokens=PARROT_CONFIGS.generative_models.chunking.max_tokens,
        temperature=PARROT_CONFIGS.generative_models.chunking.temperature,
    )

    return summaries


async def generate_final_result(
    texts: List[TimedTranscription],
    task: ParrotTask = ParrotTask.RECAP,
    use_llama_cpp: bool = False,
) -> str:
    prompt = resolve_prompt_from_task(task, language=PARROT_CONFIGS.language)
    client = get_client(use_llama_cpp)

    summaries = await generate_chunks(client, [t.text for t in texts])

    recap = await client.agenerate(
        prompt=prompt.format(texts="\n\n".join(summaries)),
        max_tokens=PARROT_CONFIGS.generative_models.text_generation.max_tokens,
        temperature=PARROT_CONFIGS.generative_models.text_generation.temperature,
    )

    return recap



"""A utility module for handling different transcript formats"""
# pylint: disable=C0103
import enum


class TranscriptType(str, enum.Enum):
    """An enum of supported transcript formats"""

    microsoftstream = "microsoftstream"
